//******************
// CHKTIMER.C
//
// this is a test of the timer
//
// The timer is installed by TimerInstall at 1000hz (1khz) (1193)
// (value is approximately 1000.151 hz)
// The system timer is called at the ratio 65536/1193, or approximately
// once each 54.934 ms, which equals the 18.2 hz it should be at
//
// The timer also provides an optional timer callback for user
// timer isr's.  The callback is called at the ratio n/1193, where n is the
// value passed when setting the callback.  In the test program the callback
// implements a simple timer with a value of 1193*10, which is exactly 
// 1/10th the timer frequency
//
// if all is well, this test shows 3 timers, all incrementing at different
// rates with a ratio of approximately 54-55 for the system timer, and
// a ratio of exactly 10 for the callback timer
//******************

#include <stdio.h>
#include <conio.h>
#include "timer.h"

#define systimer() (*(volatile unsigned int *)0x46C) // 18.2 hz

int mytimer;

// simple callback function
void callback(){
  mytimer++;
} // end callback


main(){
int ot,os,oc,nt,ns,nc,rs,rc;

TimerInstall(TIMER_1000hz);
TimerCallback(TIMER_100hz,callback);

// get initial values for all 3 timers
ot=Timer;
os=systimer();
oc=mytimer;

printf("\ntimer installed at 1000hz (1khz), ok, 1000.15088 hz\n");
printf("system ratio should be about 54-55ms = 18.2hz (54.934) - "
       "will level to 54\n");
printf("callback ratio should be exactly 10ms = 100hz\n\n");
printf("timer___  system__  ratio___  callback  ratio___\n");

while(!kbhit()){
  nt=Timer-ot;
  ns=systimer()-os;
  nc=mytimer-oc;
  if(ns!=0){rs=nt/ns;} else rs=0;
  if(nc!=0){rc=nt/nc;} else rc=0;
  printf("%8i  %8i  %8i  %8i  %8i     \r",nt,ns,rs,nc,rc);
} // end while
if(!getch()) getch();
TimerRemove();
} // end main
