//******************
// PLAY_MC.C
//
// simple .wav/.dps/raw player
// plays up to 16 samples at once
//
// Copyright (c) 1996 Manuel Suarez
// abalamahalamatandra@pobox.com
//
// wmake play_mc.exe
//
// November 11, 1996 : in practice, loading seems to limit to 9-11 channels
//                     so max cut back to 16 for better volume
// November 10, 1996 : converted from play.c
//******************

#include <stdio.h>
#include <conio.h>
#include <string.h>
#include <direct.h>
#include "src\dpam.c"

int peak_channel;

//******************
// strip the file name
//******************
void strip_name(char *string){
char *p;

p=string+strlen(string);
while(p>string && *p!='\\') p--;
if(*p=='\\') p++;
*p=0;

} // end strip_name

int LoadAndPlay(char *filename,int mode, int frequency){
  int channel,type;
  DpSample *ds;
  
  printf("%s: ",filename);
  ds=DpLoadSample(filename,&type);
  if(type) printf("%s\n",DpTypeName[type]);
  else  {printf("Not found, or too big, maybe?\n"); return 1;}
  if(mode!=-1) ds->Mode=mode;
  if(frequency!=-1) ds->Frequency=frequency;
  printf("%i bytes of %s @ %i hz\n",
    ds->Size,DpModeName[ds->Mode],ds->Frequency);
  while(!(channel=DmPlay(ds))) if(kbhit()) break;
  if(channel>peak_channel) peak_channel=channel;
  return 1;
} // end LoadAndPlay

main(int argc, char **argv){
  DIR *dp;
  int mode,frequency;
  char path[256];
  char filename[256];

  // make sure proper usage
  if(argc!=2 && argc!=4){
    printf("USAGE:  play_mc filename[.ext] [mode frequency]  (wildcards OK)\n");
    goto exit;
  } // end if

  // get mode & frequency overrides if specified
  if(argc==4){
    mode=atoi(argv[2]);
    frequency=atoi(argv[3]);
  } else {
    mode=frequency=-1;
  } // end if-else

  // install digpak & mixer
  if(!DmInstall(DM_DEFAULT,DM_DEFAULT,16)) return;

  // open a directory, if filename fails as wildcard, try as single file
  if(!(dp=opendir(argv[1]))){
    LoadAndPlay(argv[1],mode,frequency);
    goto exit;
  } // end if

  // get the path - if '.' is 1st & no wildcard, path was specified w/o '\'
  // or ending in '\', otherwise, need to strip the filename off
  strcpy(path,argv[1]);
  if(*dp->d_name=='.' &&
     !strchr(path,'*') &&
     !strchr(path,'?') &&
     path[strlen(path)-1]!='\\') strcat(path,"\\");
  else strip_name(path);

  // loop through directory entries
  for(;;){
    if(!(dp=readdir(dp))) break;
    if(dp->d_attr&_A_SUBDIR) continue;
    strcpy(filename,path);
    strcat(filename,dp->d_name);
    if(!LoadAndPlay(filename,mode,frequency)) break;
    if(kbhit()) break;
  } // end for

  closedir(dp);

  while(DmPlaying() && !kbhit());
  
exit:
  DmRemove();
  if(kbhit()) if(!getch()) getch();
  printf("\nPeak Channels Used: %i\n",peak_channel);
} // end main

