//******************
// DPAPI.C
//
// Some handy C functions for a TASM-free DIGPAK api
//
// Copyright (c) 1996 Manuel Suarez
// abalamahalamatandra@pobox.com
//
// NOTICE:  My files, DPAPI*, DPMIX* & DPAM*, are FREE SOFTWARE, you may use or
// modify them for any purpose, provided I get credit for my work
// (such as it is)
//
// DIGPAK is a product of The Audio Solution, free for non-commercial use
// with a $500 licensing fee if you sell 500 copies of a commercial work
// please read \DIGPAK\README.PRN regarding use and distribution of the
// drivers, don't take my word for it!
//
// November 10, 1996 : fixed skipping of unsuppored blocks in .wav files
//                     frees buffer DOS ram for quiet to run
// November 8, 1996 : added simple sample file format: ".dps"
// November 6, 1996 : simplfied the kludge somewhat, DpCheck in DpRemove
// November 5, 1996 : added kludge to work around problem with calling
//                    DpReportDmac from ISR under PMODE/W
// November 4, 1996 : added some support functions
// November 2, 1996 : converted from FREEMIX.C
//******************
#define DPAPI_VERSION ("1.00")

#include <stdio.h>   // file ops
#include <stdlib.h>  // getenv
#include <string.h>  // memset & strstuff
#include <i86.h>     // int386, etc
#include <process.h> // spawnl
#include "dpapi.h"

//******************
// definitions
//******************
#define SysTimer (*(volatile unsigned int *)0x46C)

//******************
// Global Variables
//******************
DpBuffer *DpBuf;
volatile unsigned short *DpFlag;
volatile unsigned short *DpSem;
volatile unsigned short *DpPending;
void *DpDMAbuf;
int DpCap;
int DpMaxMode;
static int DpPmodewFlag;
static unsigned short DpDigpakCS,DpDigpakIP;
static int DpDmaSelector;

//******************
// Strings
//******************
char *DpModeName[4]={
  "8 bit mono",
  "8 bit stereo",
  "16 bit mono",
  "16 bit stereo"
};

char *DpTypeName[4]={
  "Load failed",
  "raw",
  "DPS",
  "Wave"
};

//******************
// DpAllocateLow(size,*selector)
// allocates buffer of specified size from first 1mb of memory via DPMI
// returns the flat address
// if selector is not null, selector will be returned in *selector
// selector is needed to free memory before program terminates
//******************
void *DpAllocateLow(int size, int *selector){
  union REGS regs;
  struct SREGS sregs;

  memset(&regs,0,sizeof(regs));
  memset(&sregs,0,sizeof(sregs));
  
  regs.w.ax=0x0100;
  regs.w.bx=(size+15)>>4; //allocation is in paragraphs
  int386x(0x31,&regs,&regs,&sregs);
  if(regs.w.cflag)  return NULL;

  // ax=segment, dx=selector
  if(selector) *selector=regs.w.dx;
  memset((void *)(regs.w.ax<<4),0,size); // clear the allocated memory
  return (void *)(regs.w.ax<<4);
} // end DpAllocateLow

//******************
// DpFreeLow(selector)
// frees buffers from first 1mb of memory
//******************
int DpFreeLow(int selector){
  union REGS regs;
  struct SREGS sregs;

  memset(&regs,0,sizeof(regs));
  memset(&sregs,0,sizeof(sregs));
  
  regs.w.ax=0x0101;
  regs.w.dx=selector;
  int386x(0x31,&regs,&regs,&sregs);
  if(regs.w.cflag) return 0;

  return 1;
} // end DpFreeLow

//******************
// DpCheck
// checks to see if DIGPAK is
// installed at int 0x66
// returns 1 for yes, 0 for no
//******************
int DpCheck(void){
  union REGS regs;
  char *cp;

  regs.w.ax=0x200;
  regs.h.bl=0x66;
  int386(0x31,&regs,&regs); // check on the DIGPAK interrupt handler

  cp=(char *)((regs.w.cx<<4)+regs.w.dx);
  if(!cp) return 0;

  cp-=6;
  if(cp[0]=='K' && cp[1]=='E' && cp[2]=='R' && cp[3]=='N')
    return 1;

  return 0;
} // end DpCheck

//******************
// DpInit(segment)
// Initializes a dynamically loaded DIGPAK driver
// via DPMI call real mode function call
// This will fail under DOS/4GW
//******************
static int DpInit(int segment){
  struct SREGS sregs;
  union REGS regs;
  DpRmi rmi;
  
  memset(&rmi,0,sizeof(rmi));
  memset(&regs,0,sizeof(regs));
  memset(&sregs,0,sizeof(sregs));
  
  rmi.cs=segment;
  rmi.ip=0x200; // offset for initialize
  regs.w.ax=0x301;
  regs.h.bh=0;
  regs.w.cx=0;
  sregs.es=FP_SEG(&rmi);
  regs.x.edi=FP_OFF(&rmi);
  int386x(0x31,&regs,&regs,&sregs);

  // test for failure since function 0x301 is disabled under DOS/4GW
  if(regs.w.cflag) return 0;

//  testing here complicates reporting error messages
//  we'll just do a DpCheck afterwards
//  if(rmi.eax&0xffff) return 0;  // init failed

  return 1;
} // end DpInit

//******************
// DpUnInit(segment)
// un-initializes a dynamically loaded driver
//******************
static void DpUnInit(int segment){
  struct SREGS sregs;
  union REGS regs;
  DpRmi rmi;
    
  memset(&rmi,0,sizeof(rmi));
  memset(&regs,0,sizeof(regs));
  memset(&sregs,0,sizeof(sregs));

  rmi.cs=segment;
  rmi.ip=0x203; // offset for uninitialize
  regs.w.ax=0x301;
  regs.h.bh=0;
  regs.w.cx=0;
  sregs.es=FP_SEG(&rmi);
  regs.x.edi=FP_OFF(&rmi);
  int386x(0x31,&regs,&regs,&sregs);
} // end DpUnInit

//******************
// DpLoad(flag)
// if flag=2, will return a pointer to the driver ID text, if it was
//            dynamically loaded (I don't know how to get it from the TSR)
//            will return a msg string for TSR and for errors
//            id field is a series of strings, each ended with "\n\0"
//            and finally ended with "\n\0\0"
// if flag=1, loads and initializes soundrv.com, first testing for TSR
// if flag=0, unitializes and removes driver, if it was dynamically loaded
// (If the driver was run as a TSR it must be removed with "soundrv u")
// returns 1 for OK, 0 for fail.
//
// DYNAMIC LOADING REQUIRES PMODE/W!
// Royalty free DOS/4GW has DMPI service $301 disabled. PMODE/W supports it
// DOS/4GW won't crash, it just won't init the driver and that is detected
// This means the TSR method must be used with DOS/4GW
//
// Note that the DIGAUTO driver does not support dynamic loading and
// must always be used as a TSR.
//
// DIGPAK drivers below version 3.4 will not work in protected mode, so
// those are also rejected by this routine
//
// a "DIGPAK" environment variable is supported for locating the driver
// set to full path e.g.:  set digpak=c:\utility\soundrv.com
//******************
int DpLoad(int flag){
  FILE *fp;
  int i,size;
  char *cp;
  static char dp[]="DIGPAK";

  static char NoDriverMsg[]="No driver loaded\n\0";
  static char TsrMsg[]="DIGPAK TSR is installed\n\0";
  static char InvalidMsg[]="INVALID DRIVER\n\0";
  static char Dos4gwMsg[]=
    "INIT CALL FAILED (DOS/4GW does not support) - use TSR\n\0";
  static char FailedMsg[]=
    "DRIVER INIT FAILED (wrong driver or sound was playing in Win95?)\n\0";
  static char OldMsg[]="DRIVER VERSION BELOW 3.40\n\0";
  static char AutoMsg[]="DIGAUTO CANNOT BE DYNAMICALLY LOADED\n\0";
  static char *DpText=NoDriverMsg;
  static int DpLoaded=0;

  // See if a request for the id field is being made
  if(flag==2)  return (int)DpText;

  // Dynamically load and initialize DIGPAK driver, if TSR is not present
  if(flag){

  // first see if there is already a driver installed
  // and if it is proper version
  if(DpCheck()){
    if(!DpLoaded) DpText=TsrMsg; // I don't know how to get id from TSR
    if(DpReportVersionNumber()<340) return 0; // no go in protected mode
    return 1;
  } // end if

  // check for driver in current dir
  // then check for "digpak" environment variable
  if(!(fp=fopen("soundrv.com","rb"))){
    if(getenv("digpak")){
      if(!(fp=fopen(getenv("digpak"),"rb"))) return 0;
    } else return 0;
  } // end if

  // get file length, allocate memory & read it in
  fseek(fp,0,SEEK_END);
  size=ftell(fp);
  fseek(fp,0,SEEK_SET);
  cp=DpAllocateLow(size,NULL);
  if(!cp){
     fclose(fp);
     return 0;
  } // end if
  i=fread(cp,1,size,fp);
  fclose(fp);
  if(i!=size){
    return 0;
  } // end if

  // verify it's a digpak driver
  // also rejects DIGAUTO.COM since it can't be dynamically initilized
  for(i=0;i<6;i++)
    if(cp[i+3]!=dp[i]){
      if(cp[i+3]=='A' && cp[i+4]=='U' && cp[i+5]=='T' && cp[i+6]=='O'){
         DpText=AutoMsg;
         return 0;
       } else {
         DpText=InvalidMsg;
         return 0;
       } // end else
    } // end if
     
  // initialize driver
  DpLoaded=((int)cp>>4)-0x10; // get segment & adjust to proper position
  if(!DpInit(DpLoaded)){
    DpLoaded=0;
    DpText=Dos4gwMsg;
    return 0;
  } // end if
  
  // make sure it really initialized
  if(!DpCheck()){
    DpLoaded=0;
    DpText=FailedMsg;
    return 0;
  } // end if

  // make sure it is proper version
  if(DpReportVersionNumber()<340){
     DpUnInit(DpLoaded); // kill it
     DpText=OldMsg;
     return (DpLoaded=0);
  } // end if

  // set the id field address
  DpText=cp+3;

  // hey, remember that if(flag)?
  // we need to uninitilize the driver
  } else {

  // make sure we initialized it before trying to uninitialize it
  if(!DpLoaded) return 0;

  // now un-init it
  DpUnInit(DpLoaded);
  DpLoaded=0;
  DpText=NoDriverMsg;

  } // end if-else
  
  return 1;
} // end DpLoad

//******************
// DpPrintId()
// outputs the driver id string to stdout
//******************
void DpPrintId(void){
  char *cp;

  cp=(char *)DpLoad(2);
  while(*cp){
    printf(cp);
    while(*cp++);
  } // end while

} // end DpPrintId

//******************
// DpRunLoud()
// runs the "loud" environment variable
// e.g. set loud=c:\sb16\mixerset /ma:255 /q
//******************
void DpRunLoud(void){
  int i;
  char str[256];

  if(getenv("loud")){
    strcpy(str,getenv("loud"));
    for(i=0;str[i]!=' ';i++);
    str[i++]=0;
    spawnl(P_WAIT,str," ",str+i,NULL);
  } // end if
} // end DpRunLoud

//******************
// DpRunQuiet()
// runs the "quiet" environment variable
// e.g. set quiet=c:\sb16\mixerset /p /q
//******************
void DpRunQuiet(void){
  int i;
  char str[256];

  if(getenv("quiet")){
    strcpy(str,getenv("quiet"));
    for(i=0;str[i]!=' ';i++);
    str[i++]=0;
    spawnl(P_WAIT,str," ",str+i,NULL);
  } // end if
} // end DpRunQuiet

//******************
// DpPause(ticks)
// pauses specified # of timer ticks
//******************
void DpPause(int ticks){
  int stop;
  
  stop=SysTimer+ticks;
  while(SysTimer<stop);
} // end DpPause

//******************
// the DMAC kludge
// These functions are used to implement a sloppy kludge to work around
// a bizarre problem under PMODE/W.  PMODE/W apparently does not like
// the int 0x66 call to get the DMA counter to be made during the service
// interrupt.  What happens is that kbhit() falsely reports that a key
// has been pressed, thus a line like: if(kbhit()) k=getch();  will stop
// everything until a key is really pressed.  This kludge gets around the
// problem by retrieving the real mode interrupt vector for 0x66.
// if PMODE/W is the extender and there is not another DPMI host (e.g. win95)
// then DPMI function 302 (call real mode procedure with an interrupt frame)
// is used to call DIGPAK.  Otherwise, the standard int 0x66 call is made.
// This is rather sloppy, but it did solve the problem.
// another fix, by the way, is to use a custom keyboard handler.  A simple
// one is provided as KEY.C
// To use:
//   This specifically relates to a DpReportDmac call from an ISR
//   The problem does not appear if DpReportDmac is called elsewhere
//   If certain PMODE/W won't be used, just use DpReportDmac
//   otherwise, replace DpReportDmac() with DpDmacKludge()
//   DpInstall() handles the setup
//******************
extern int IsPmodew();
#pragma aux IsPmodew= \
  "mov eax,0xeeff" \
  "int 0x31" \
  "cmp eax,'PMDW'" \
  "jne no" \
  "cmp ch,3" /* if another DPMI (Win95) is providing services, all is OK*/ \
  "je no" \
  "mov eax,1" \
  "jmp yes" \
  "no:" \
  "mov eax,0" \
  "yes:" \
  value [eax];

extern void GetDigpakRmVect(void);
#pragma aux GetDigpakRmVect= \
  "mov eax,0x200" \
  "mov bl,0x66" \
  "int 0x31" \
  "mov DpDigpakCS,cx" \
  "mov DpDigpakIP,dx" \
  modify [eax ebx ecx edx];

extern int KludgeProcCall(DpRmi *);
#pragma aux KludgeProcCall= \
  "mov eax,0x302" \
  "xor ebx,ebx" \
  "xor ecx,ecx" \
  "int 0x31" \
  modify [ebx ecx] parm [edi] value [eax];

#ifndef __SW_S
#pragma off (check_stack)
#endif
int DpDmacKludge(void){
  DpRmi rmi;

  if(!DpPmodewFlag) return DpReportDmac();

  rmi.ss=rmi.sp=0;
  rmi.eax=0x69d;
  rmi.cs=DpDigpakCS;
  rmi.ip=DpDigpakIP;
  KludgeProcCall(&rmi);

  return rmi.eax;
} // end DpDmacKludge
#ifndef __SW_S
#pragma on (check_stack)
#endif

//******************
// DpInstall()
// Installs DIGPAK and initializes data structures
// Note, this will output some text to STDOUT
// return 1 for OK
//        0 for FAILURE
//******************
int DpInstall(void){
  int i;

  // display a little banner
  printf("DPAPI %s (c)1996 Manuel Suarez abalamahalamatandra@pobox.com\n\n",
         DPAPI_VERSION);
  
  // pump up the volume
  DpRunLoud();

  // do we have a valid driver?
  i=DpLoad(1);

  // display the driver ID or error message
  DpPrintId();
  printf("\n");
  if(!i) return 0;

  // gather info for the DMAC kludge to work around the PMODE/W kbhit prob.
  DpPmodewFlag=IsPmodew();
  GetDigpakRmVect();

  // set protected mode interface
  DpSetDPMIMode();

  // get driver capabilities & determine maximum mode
  DpCap=DpAudioCapabilities();
  if(DpCap&DP_16S) DpMaxMode=DP_MODE_16S;
  else if(DpCap&DP_16M) DpMaxMode=DP_MODE_16M;
  else if(DpCap&DP_8S)  DpMaxMode=DP_MODE_8S;
  else DpMaxMode=DP_MODE_8M;

  // allocate the DIGPAK structure
  if(!(DpBuf=DpAllocateLow(sizeof(DpBuffer),NULL))) return 0;

  // allocate the flag
  if(!(DpFlag=DpAllocateLow(2,NULL))) return 0;

  // allocate 2x the buffer, save selector so we can free mem for DpRunQuiet
  if(!(DpDMAbuf=DpAllocateLow(DP_DMA_BUFSIZE*2,&DpDmaSelector))) return 0;

  // test buffer for validity, and use 2nd half if needed
  if(!DpVerifyDMA(DpDMAbuf,DP_DMA_BUFSIZE))
    DpDMAbuf=(char *)DpDMAbuf+DP_DMA_BUFSIZE;

// alternate test, (a little messier)
//  if((((int)DpDMAbuf&0xffff)+DP_DMA_BUFSIZE)>65536)
//    DpDMAbuf=(char *)DpDMAbuf+DP_DMA_BUFSIZE;

  // get semaphore and pending addresses
  DpSem=DpReportSemaphoreAddress();
  DpPending=DpReportPendingAddress();

  // set default playback mode
  DpSetPlayMode(DP_MODE_8M);

  // initialize DpBuffer with default stuff
  DpBuf->BufOffset=0;
  DpBuf->BufSegment=(int)DpDMAbuf>>4;
  DpBuf->Length=DP_DMA_BUFSIZE;
  DpBuf->StatusOffset=0;
  DpBuf->StatusSegment=(int)DpFlag>>4;
  DpBuf->Frequency=11025;

  return 1;
} // end DpInstall

//******************
// DpRemove()
// uninitializes and removes DIGPAK
//******************
void DpRemove(void){
  // See if there is something to remove
  if(!DpCheck()) return;

  // stop any sound that's playing
  // w/o pause, DMA could be killed until driver is reloaded
  DpStopSound();
  DpPause(4);

  // unload and un-init if driver was dynamically loaded
  DpLoad(0);

  // free buffer memory to make room for DpRunQuiet
  DpFreeLow(DpDmaSelector);
 
  // quiet it back down
  DpRunQuiet();
} // end DpRemove

//*****************
// DpOn(mode,frequency,buffer size)
// starts playing in DMA autoinit mode using specified settings
// buffer size should be <=DP_DMA_HALFBUF & a multiple of 16
// autoinit buffer will be twice this value
// NOTE:  assumes DpInstall has been called
// returns 1 for OK
//         0 for FAIL
// It appears that, at least for the sb16 driver, DpDigPlay
// can be called with new settings while a sample is playing
//*****************
int DpOn(int mode,int frequency,int bufsize){
  // See if there is something to turn on
  if(!DpCheck()) return 0;
  
  if(!DpSetPlayMode(mode)) return 0;
  if(!DpSetBackFillMode(1)) return 0;
  DpBuf->Length=bufsize*2;
  DpBuf->Frequency=frequency;

  DpDigPlay(DpBuf);

  return 1;
} // end DpOn

//******************
// DpOff()
// stops safely
//******************
void DpOff(void){
  DpStopSound();
  DpPause(4);
} // end DpOff

//******************
// DpClearDMAbuf(mode)
// clears the DMA buffer as appropriate for the given mode
// this will prevent an initial click when starting to play in 8 bit modes
// with a zeroed buffer (baseline in 8 bit is 0x80, not 0)
//******************
void DpClearDMAbuf(int mode){
  if(mode==DP_MODE_8M || mode==DP_MODE_8S)
    memset(DpDMAbuf,0x80,DP_DMA_BUFSIZE);
  else
    memset(DpDMAbuf,0,DP_DMA_BUFSIZE);
} // end DpClearDMAbuf

//******************
// DpLoadSample(filename,*type)
// tries filename in this order:
// .dps, .wav, raw
// allocates memory and loads filename into it
// returns pointer to a DpSample structure (also allocated) or NULL
// *type set as follows if not NULL
//         0 for fail
//         1 for raw format
//         2 for dps format
//         3 for wav format
//******************
DpSample *DpLoadSample(char *filename,int *type){
  DpSample *ds;
  int t;

  if(ds=DpLoadDps(filename)) t=2;
  else if(ds=DpLoadWav(filename)) t=3;
  else if(ds=DpLoadRaw(filename)) t=1;
  else t=0;
 
  if(type) *type=t;
  return ds;
} // end DpLoadSample

//******************
// DpSaveSample(filename,sample,type)
// Saves filename as specified type:
//     1 for raw format
//     2 for dps format
//     3 for wav format
// returns 1 for OK, 0 for fail
//******************
int DpSaveSample(char *filename,DpSample *ds,int type){
  switch(type){
    case 1: return DpSaveRaw(filename,ds);
    case 2: return DpSaveDps(filename,ds);
    case 3: return DpSaveWav(filename,ds);
  } // end switch

  return 0;
} // end DpSaveSample

//******************
// DpFreeSample(sample)
// presumes both the sample structure
// and the buffer it points to
// were allocated by malloc.
//******************
void DpFreeSample(DpSample *sample){
  free(sample->Buffer);
  free(sample);
} // end DpFreeSample

//******************
// DpLoadRaw(filename)
// allocates memory and loads filename into it
// returns pointer to a DpSample structure (also allocated)
//******************
DpSample *DpLoadRaw(char *filename){
  FILE *fp;
  int i,size;
  DpSample *ds;
  void *buf;
    
  // open the file
  fp=fopen(filename,"rb");
  if(!fp) return NULL;
  
  // get the size of the file
  fseek(fp,0,SEEK_END);
  size=ftell(fp);
  fseek(fp,0,SEEK_SET);

  // allocate buffer & sample structure
  buf=malloc(size);
  if(!buf){fclose(fp); return NULL;}
  ds=malloc(sizeof(DpSample));
  if(!ds){fclose(fp); free(buf); return NULL;}

  // set the DpSample structure
  ds->Mode=DP_MODE_8M;  // no mode info, so just set a default
  ds->Frequency=11025;  // same here
  ds->Size=size;
  ds->Buffer=buf;
  
  // read it in
  i=fread(buf,1,size,fp);
  fclose(fp);

  // did it work?
  if(i==size) return ds;

  // ugh, failure!
  free(buf);
  free(ds);
  return NULL;
} // end DpLoadRaw

//******************
// DpSaveRaw(filename,sample)
// Saves sample as raw data
// returns 1 for success, or 0 for failure
//******************
int DpSaveRaw(char *filename,DpSample *ds){
  FILE *fp;
  int i;
    
  // open the output file
  fp=fopen(filename,"wb");
  if(!fp) return 0;
  
  // write it out
  i=fread(ds->Buffer,1,ds->Size,fp);
  fclose(fp);

  // did it work?
  if(i==ds->Size) return 1;

  // failure!
  return 0;
} // end DpSaveRaw

//******************
// DpLoadDps(filename)
// allocates memory and loads filename into it
// ".dps" files are described in DpSaveDps below
// returns pointer to a DpSample structure (also allocated)
//******************
DpSample *DpLoadDps(char *filename){
  FILE *fp;
  DpDpsHeader dps;
  DpSample *ds;
  int i;
  char *buf;
    
  // set ds to NULL so error routine will work
  ds=NULL;
  
  // open the file, if no extension ".dps" is added
  buf=malloc(strlen(filename)+8);
  if(!buf) return NULL;
  strcpy(buf,filename);
  for(i=0;buf[i];i++) if(buf[i]=='.') break;
  if(!buf[i]) strcat(buf,".dps");
  fp=fopen(buf,"rb");
  free(buf);
  buf=NULL;
  if(!fp) return NULL;
  
  // read in the .dps header
  if(fread(&dps,1,sizeof(dps),fp)!=sizeof(dps)) goto error;

  // check tag and version
  if(dps.Tag!='DPSF' || dps.Version!=DP_DPS_VERSION) goto error;
  
  // allocate buffer & sample structure
  buf=malloc(dps.Size);
  if(!buf) goto error;
  ds=malloc(sizeof(DpSample));
  if(!ds) goto error;

  // set the DpSample structure
  ds->Mode=dps.Mode;
  ds->Frequency=dps.Frequency;
  ds->Size=dps.Size;
  ds->Buffer=buf;
  
  // read it in
  i=fread(buf,1,dps.Size,fp);

  // did it work?
  if(i==dps.Size){
    fclose(fp);
    return ds;
  } // end if

  // ugh, failure!
error:
  if(buf) free(buf);
  if(ds)  free(ds);
  fclose(fp);
  return NULL;
} // end DpLoadDps

//******************
// DpSaveDps(filename,sample)
// Saves sample as a ".dps" file
// format:  'DPSF',version,mode,frequency,size-of-data,data
// current version is 1
// returns 1 for success, or 0 for failure
//******************
int DpSaveDps(char *filename,DpSample *ds){
  FILE *fp;
  DpDpsHeader dps;
  int i;
  char *buf;
    
  // open the output file, first ensuring a ".dps" extension
  buf=malloc(strlen(filename)+8);
  if(!buf) return 0;
  strcpy(buf,filename);
  for(i=0;buf[i];i++) if(buf[i]=='.') break;
  buf[i]=0;
  strcat(buf,".dps");
  fp=fopen(buf,"wb");
  free(buf);
  buf=NULL;
  if(!fp) return 0;

  // build and write the header
  dps.Tag='DPSF';
  dps.Version=DP_DPS_VERSION;
  dps.Mode=ds->Mode;
  dps.Frequency=ds->Frequency;
  dps.Size=ds->Size;
  if(fwrite(&dps,1,sizeof(dps),fp)!=sizeof(dps)) goto error;

  // write the data
  i=fwrite(ds->Buffer,1,ds->Size,fp);

  // did it work?
  if(i==ds->Size){
    fclose(fp);
    return 1;
  } // end if

  // failure!
error:
  fclose(fp);
  return 0;
} // end DpSaveDps

//******************
// DpLoadWav(filename)
// if no extension, adds ".wav"
// allocates memory and loads filename into it
// returns pointer to a DpSample structure (also allocated)
// (p.s. this format sucks!!)
//******************
DpSample *DpLoadWav(char *filename){
  FILE *fp;
  int i,wavemode,wavesize,filesize,nextseek;
  WaveFormatChunk wf;
  DpSample *ds;
  char *buf;

  // set ds to NULL so error routine will work
  ds=NULL;
  
  //******************
  // open the wave file
  // if no extension, adds ".wav"
  //******************
  buf=malloc(strlen(filename)+8);
  if(!buf) return NULL;
  strcpy(buf,filename);
  for(i=0;buf[i];i++) if(buf[i]=='.') break;
  if(!buf[i]) strcat(buf,".wav");
  fp=fopen(buf,"rb");
  free(buf);
  buf=NULL;
  if(!fp) return NULL;

  //******************
  // really a RIFF?
  //******************
  if(fread(&i,1,4,fp)!=4) goto error;
  if(i!=WAV_RIFFID) goto error;

  //******************
  // filesize is RIFF size + header size
  //******************
  if(fread(&filesize,1,4,fp)!=4) goto error;
  filesize+=ftell(fp);

  //******************
  // really a WAVE?
  //******************
  if(fread(&i,1,4,fp)!=4) goto error;
  if(i!=WAV_WAVEID) goto error;

  //****************
  // get the current file pointer
  //****************
  nextseek=ftell(fp);

  //******************
  // skip everything
  // which is not a format chunk
  //******************
  if(fread(&i,1,4,fp)!=4) goto error;
  nextseek+=4;
  while(i!=WAV_FMTID){
    fread(&i,1,4,fp);
    nextseek+=i+4;
    if(nextseek>=filesize) goto error;
    fseek(fp,nextseek,SEEK_SET);
    if(fread(&i,1,4,fp)!=4) goto error;
    nextseek+=4;
  } // end while

  //******************
  // get the format chunk & make sure we've skipped all of it
  //******************
  if(fread(&i,1,4,fp)!=4) goto error;
  nextseek+=i+4;
  if(fread(&wf,1,sizeof(wf),fp)!=sizeof(wf)) goto error;
  fseek(fp,nextseek,SEEK_SET);

  //******************
  // now find the associated data
  //******************
  if(fread(&i,1,4,fp)!=4) goto error;
  nextseek+=4;
  while(i!=WAV_DATAID){
    if(fread(&i,1,4,fp)!=4) goto error;
    nextseek+=i+4;
    if(nextseek>=filesize) goto error;
    fseek(fp,nextseek,SEEK_SET);
    if(fread(&i,1,4,fp)!=4) goto error;
    nextseek+=4;
  } // end while

  //******************
  // The size of the data
  // followed by the data
  // it's what we've been wating for!
  //******************
  if(fread(&wavesize,1,4,fp)!=4) goto error;
  nextseek+=4;

  //******************
  // check the tag for PCM file id
  //******************
  if(wf.Tag!=1) goto error;

  //******************
  // check the channels (1=mono, 2=stereo)
  //******************
  switch(wf.Channels){
  case 1:
    wavemode=0;
    break;
  case 2:
    wavemode=1;
    break;
  default:
    goto error;
  } // end switch
  
  //******************
  // check the bits
  //******************
  switch(wf.Bits){
  case 8:
    break;
  case 16:
    wavemode+=2;
    break;
  default:
    goto error;
  } // end switch

  // allocate a buffer & sample structure
  buf=malloc(wavesize);
  if(!buf) goto error;
  ds=malloc(sizeof(DpSample));
  if(!ds) goto error;

  // read in the data
  i=fread(buf,1,wavesize,fp);
  if(i!=wavesize) goto error;
  
  // done!
  fclose(fp);
  ds->Mode=wavemode;
  ds->Frequency=wf.SampleRate;
  ds->Size=wavesize;
  ds->Buffer=buf;
  return ds;

error:
  if(ds) free(ds);
  if(buf) free(buf);
  fclose(fp);
  return NULL;
} // end DpLoadWav

//******************
// DpSaveWav(filename,sample)
// ".wav" extension will be added if not already there
// returns 1 for OK, 0 for failed
//******************
int DpSaveWav(char *filename,DpSample *ds){
  FILE *fp;
  int i;
  WaveFormatChunk wf;
  char *buf;

  //******************
  // open the output file, first ensuring a ".wav" extension
  //******************
  buf=malloc(strlen(filename)+8);
  if(!buf) return 0;
  strcpy(buf,filename);
  for(i=0;buf[i];i++) if(buf[i]=='.') break;
  buf[i]=0;
  strcat(buf,".wav");
  fp=fopen(buf,"wb");
  free(buf);
  buf=NULL;
  if(!fp) return 0;

  //******************
  // write RIFF id
  //******************
  i=WAV_RIFFID;
  if(fwrite(&i,1,4,fp)!=4) goto error;

  //******************
  // RIFF size is data size + header size
  // header doesn't include RIFFID and this field
  //******************
  i=ds->Size+sizeof(wf)+20;
  if(fwrite(&i,1,4,fp)!=4) goto error;

  //******************
  // write WAVE id
  //******************
  i=WAV_WAVEID;
  if(fwrite(&i,1,4,fp)!=4) goto error;

  //******************
  // write FORMAT id
  //******************
  i=WAV_FMTID;
  if(fwrite(&i,1,4,fp)!=4) goto error;

  //******************
  // write the format chunk
  //******************
  i=sizeof(WaveFormatChunk); // length of chunk
  if(fwrite(&i,1,4,fp)!=4) goto error;
  i=1; if(ds->Mode&1) i*=2; if(ds->Mode&2)i*=2;  // i=bytes per sample
  wf.Tag=1; // PCM file
  wf.Channels=(ds->Mode&1 ? 2 : 1); // mono=1, stereo=2
  wf.SampleRate=ds->Frequency; // samples per second
  wf.ByteRate=ds->Frequency*i; // bytes per second
  wf.Block=i; // block alignment (bytes per sample)
  wf.Bits=(ds->Mode&2 ? 16 : 8); // bits per sample
  if(fwrite(&wf,1,sizeof(wf),fp)!=sizeof(wf)) goto error;

  //******************
  // now write the DATA id
  //******************
  i=WAV_DATAID;
  if(fwrite(&i,1,4,fp)!=4) goto error;

  //******************
  // Write out the size of the data
  // followed by the data
  //******************
  i=ds->Size;
  if(fwrite(&i,1,4,fp)!=4) goto error;
  if(fwrite(ds->Buffer,1,ds->Size,fp)!=ds->Size) goto error;

  // done!
  fclose(fp);
  return 1;

error:
  fclose(fp);
  return 0;
} // end DpSaveWav

