//******************
// "DPMIX.C"
//
// Copyright (c) 1996 Manuel Suarez
// abalamahalamatandra@pobox.com
//
// Multichannel Digital Audio Mixer using DPAPI to access
// The Audio Solution's DIGPAK drivers
//
// Currently requires DMA auto init, i.e. does not support old cards
//
// (note to myself)  For effects, change i=DmBufSamples to only enough
// samples for 1 timing unit.  Loop through until the i's add up to
// DmBufSamples.  In each loop, check for effects.  Maybe put an effect
// pointer in the structure.  Would help if we can assure DmBufSamples
// is always an even multiple of timing units.
//
// November 5-9, 1996 : initial version
//******************
#define DPMIX_VERSION ("0.01")

#include <stdio.h>
#include <string.h>
#include "dpam.h"

//******************
// Channel structure
//******************
typedef struct _DmChannel{
  int Active;       // Flag for whether the channel is active
  int Mode;         // Mode of the data
  int Frequency;    // Playback frequency
  int Size;         // size in bytes of the data
  void *Buffer;     // address of start of data
  int Vol;          // base volume for the channel
  int Bal;          // base balance for the channel
  int LeftVol;      // Volume for the left channel
  int RightVol;     // Volume for the right channel
  int Fstep;        // base step size for frequency interpolation
  int Fadj;         // adjustment for frequency interpolation
  int Ferr;         // error term for frequency interpolation
  int Count;        // bytes remaining in data
  void *Pos;        // address of current position within data
} DmChannel;

//******************
// global variables for talking to the ISR
//******************
static int DmMode;       // mixer output mode
static int DmFrequency;  // mixer output frequency
static int DmBufsize;    // mixer output buffer size
static int DmBufSamples; // mixer output buffer size in DmMode samples
static int DmSem;        // semaphore to disable the mixer
static int DmTemp[DP_DMA_BUFSIZE]; // temporary 32 bit mix buffer
static int DmVol,DmBal; // master volume & balance
static int DmLeftVol,DmRightVol; // master volume, balance adjusted
static int DmChannels;  // number of channels
static DmChannel DmC[DM_MAX_CHANNELS];  // the channels

static short DmSS;
static int DmESP;
static char DmStack[1024];

//******************
// The timer callback, in fact, the whole mixer
//******************
#ifndef __SW_S
#pragma off (check_stack)
#endif
void DmCallback(void){
  static int WantFirst=0;
  int i,c,err,count;
  char *cp;
  int *ip;

  if(DmSem) return;
  DmSem=1;

  // determine which buffer we are in and whether the one we want is free
  if(DpDmacKludge()>=DmBufsize){
    if(WantFirst) goto exit;  // first is unavailable
  } else {
    if(!WantFirst) goto exit;  // second is unavalable
  } // end if-else
  memset(DmTemp,0,DmBufSamples<<3); // prep the temporary buffer

  for(c=0;c<DmChannels;c++)
  if(DmC[c].Active){
    ip=DmTemp;
    cp=DmC[c].Pos;
    i=DmBufSamples;
    count=DmC[c].Count;
    err=DmC[c].Ferr;

    // upsampling, skip through source, error reset is DmFrequency
    if(DmC[c].Frequency>=DmFrequency){

    switch(DmC[c].Mode){

    case DP_MODE_8M:
      while(i--){
        *ip+=((*cp-128)<<8)*DmC[c].LeftVol;  // convert from 8m
        ip++;
        *ip+=((*cp-128)<<8)*DmC[c].RightVol; // use same source again
        ip++;

        cp+=DmC[c].Fstep;
        count-=DmC[c].Fstep;
        if((err-=DmC[c].Fadj)<=0){
          cp++;
          count--;
          err+=DmFrequency;
        } // end if
        if(count<=0) break;

      } // end while
      break;

    case DP_MODE_8S:
      while(i--){
        *ip+=((*cp-128)<<8)*DmC[c].LeftVol;  // convert from 8s
        ip++;
        *ip+=((*(cp+1)-128)<<8)*DmC[c].RightVol;
        ip++;

        cp+=DmC[c].Fstep;
        count-=DmC[c].Fstep;
        if((err-=DmC[c].Fadj)<=0){
          cp+=2;
          count--;
          err+=DmFrequency;
        } // end if
        if(count<=0) break;
      } // end while
      break;

    case DP_MODE_16M:
      while(i--){
        *ip+=*(short *)cp*DmC[c].LeftVol;  // convert from 16m
        ip++;
        *ip+=*(short *)cp*DmC[c].RightVol;  // use same source again
        ip++;
        
        cp+=DmC[c].Fstep;
        count-=DmC[c].Fstep;
        if((err-=DmC[c].Fadj)<=0){
          cp+=2;
          count-=2;
          err+=DmFrequency;
        } // end if
        if(count<=0) break;
        
      } // end while
      break;

    case DP_MODE_16S:
      while(i--){
        *ip+=*(short *)cp*DmC[c].LeftVol;  // convert from 16s
        ip++;
        *ip+=*(short *)(cp+2)*DmC[c].RightVol;
        ip++;

        cp+=DmC[c].Fstep;
        count-=DmC[c].Fstep;
        if((err-=DmC[c].Fadj)<=0){
          cp+=4;
          count-=4;
          err+=DmFrequency;
        } // end if
        if(count<=0) break;
      } // end while
      break;
    
    } // end switch

    //-----------------------------------------------------------------
    // downsampling, repeat source, error reset is DmFrequency (w/o rpt)
    } else {

    switch(DmC[c].Mode){

    case DP_MODE_8M:
      while(i--){
        *ip+=((*cp-128)<<8)*DmC[c].LeftVol;  // convert from 8m
        ip++;
        *ip+=((*cp-128)<<8)*DmC[c].RightVol;
        ip++;

        if((err-=DmC[c].Fadj)<=0){
          cp++;
          count--;
          err+=DmFrequency;
          if(count<=0) break;
        } // end if
        
      } // end while
      break;

    case DP_MODE_8S:
      while(i--){
        *ip+=((*cp-128)<<8)*DmC[c].LeftVol;  // convert from 8s
        ip++;
        *ip+=((*(cp+1)-128)<<8)*DmC[c].RightVol;
        ip++;

        if((err-=DmC[c].Fadj)<=0){
          cp+=2;
          count-=2;
          err+=DmFrequency;
          if(count<=0) break;
        } // end if
      } // end while
      break;

    case DP_MODE_16M:
      while(i--){
        *ip+=*(short *)cp*DmC[c].LeftVol;  // convert from 16m
        ip++;
        *ip+=*(short *)cp*DmC[c].RightVol;
        ip++;

        if((err-=DmC[c].Fadj)<=0){
          cp+=2;
          count-=2;
          err+=DmFrequency;
          if(count<=0) break;
        } // end if
        
      } // end while
      break;

    case DP_MODE_16S:
      while(i--){
        *ip+=*(short *)cp*DmC[c].LeftVol;  // convert from 16s
        ip++;
        *ip+=*(short *)(cp+2)*DmC[c].RightVol;
        ip++;

        if((err-=DmC[c].Fadj)<=0){
          cp+=4;
          count-=4;
          err+=DmFrequency;
          if(count<=0) break;
        } // end if
        
      } // end while
      break;
    
    } // end switch

    } // end if frequency

    DmC[c].Ferr=err;
    DmC[c].Pos=cp;
    DmC[c].Count=count;
    if(DmC[c].Count<=0) DmC[c].Active=0;
  } // end if active & for channels

  // copy temp buffer DMA buffer, converting to appropriate output mode
  ip=DmTemp;
  cp=DpDMAbuf;  if(!WantFirst) cp+=DmBufsize;
  i=DmBufSamples;
  switch(DmMode){
    
  case DP_MODE_8M:
    while(i--){
      *cp=((*ip>>1)+(*(ip+1)>>1)>>24)+128;  // convert to mono 8
      cp++; ip+=2;
    } // end while
    break;
    
  case DP_MODE_8S:
    while(i--){
      *cp=(*ip>>24)+128;  // convert to stereo 8
      cp++; ip++;
      *cp=(*ip>>24)+128;
      cp++; ip++;
    } // end while
    break;
    
  case DP_MODE_16M:
    while(i--){
      *(short *)cp=((*ip>>1)+(*(ip+1)>>1)>>16);  // convert to mono 16
      cp+=2; ip+=2;
    } // end while
    break;
    
  case DP_MODE_16S:
    while(i--){
      *(short *)cp=*ip>>16;  // convert to stereo 16
      cp+=2; ip++;
      *(short *)cp=*ip>>16;
      cp+=2; ip++;
    } // end for
    break;
    
  } // end switch
  
  // toggle buffer wanted
  WantFirst=!WantFirst;

exit:
  DmSem=0;
} // end DmCallback
#ifndef __SW_S
#pragma off (check_stack)
#endif

//******************
// The mixer needs more stack space than DPMI host is providing
// (this really should be moved into the timer)
//******************
void SetStack(void);
#pragma aux SetStack= \
"mov DmSS,ss" \
"mov DmESP,esp" \
"mov ax,ds" \
"mov ss,ax" \
"mov esp,offset DmStack+1024" \
modify [eax];

void ResetStack(void);
#pragma aux ResetStack= \
"mov ss,DmSS" \
"mov esp,DmESP";

#ifndef __SW_S
#pragma off (check_stack)
#endif
void DmCallbackCaller(void){
  SetStack();
  DmCallback();
  ResetStack();
} // end DmCallbackCaller
#ifndef __SW_S
#pragma on (check_stack)
#endif

//******************
// DmInstall(mode,frequency,channels)
// installs DIGPAK, the timer and the callback
// if mode is DM_DEFAULT, the maximum supported mode will be used
// if frequency is DM_DEFAULT, 44100 hz will be used
// if channels is DM_DEFAULT, DM_MAX_CHANNELS will be used
//******************
int DmInstall(int mode,int frequency,int channels){
  // initialize the channels (need to be certain Active=0 for all)
  memset(DmC,0,sizeof(DmC));

  // set number of channels
  if(channels<1 || channels>DM_MAX_CHANNELS) DmChannels=DM_MAX_CHANNELS;
  else DmChannels=channels;

  // initialize volume & balance
  DmLeftVol=DmRightVol=DmVol=DM_MAX_VOL/DmChannels;
  DmBal=0;

  // install DIGPAK, timer & timer callback  (0=18.2 hz)
  printf("DPMIX %s, ",DPMIX_VERSION);
  if(!DpInstall()) return 0;
  TimerInstall(DM_TIMER_FREQ);
  TimerCallback(DM_CALLBACK_FREQ,DmCallbackCaller);

  // check if defaults requested
  if(mode==DM_DEFAULT) mode=DpMaxMode;
  if(frequency==DM_DEFAULT) frequency=44100;

  // calculate a minimum buffer size for an 18.2 hz service
  // a single large size could be chosen to cover all possibilities
  // (i.e. 16s@44100), but doing this will minimize latencey
  DmBufSamples=DmBufsize=
    ((((unsigned)frequency*DM_CALLBACK_FREQ/TIMER_FREQ)+15)&~0x0f);
  if(mode&1) DmBufsize*=2; // double for stereo
  if(mode&2) DmBufsize*=2; // double for 16 bits
  if(DmBufsize*2<=DP_DMA_HALFBUF){DmBufsize*=2; DmBufSamples*=2;}
  DmMode=mode;
  DmFrequency=frequency;
  
  // clear buffer to appropriate baseline - 8 bit is 0x80, not 0x00
  DpClearDMAbuf(DmMode);

  // get DIGPAK to start looping the buffer in autoinit DMA mode
  if(!DpOn(DmMode,DmFrequency,DmBufsize)) goto error;

  printf("Mixer installed at %i hz, %s  (Timer: %0.1fhz / %0.1fhz)\n\n",
    DmFrequency,DpModeName[DmMode],
    (double)TIMER_FREQ/DM_TIMER_FREQ,(double)TIMER_FREQ/DM_CALLBACK_FREQ);

  return 1;
      
error:
  TimerRemove();
  DpRemove();
  return 0;
} // end DmInstall

//******************
// DmRemove()
// removes DIGPACK and the timer
//******************
void DmRemove(void){
  // remove the timer 1st, since the callback may be calling DIGPAK
  TimerRemove();

  // now remove DIGPAK
  DpRemove();
} // end DmRemove

//******************
// DmUpdateVolume(channel)
// internal subroutine for updating volume values
//******************
static void DmUpdateVolume(int c){
  DmC[c].LeftVol=(DmC[c].Vol*DmLeftVol)>>15;
  DmC[c].RightVol=(DmC[c].Vol*DmRightVol)>>15;
  if(DmC[c].Bal<0)
      DmC[c].RightVol=(DmC[c].RightVol*(DM_MAX_VOL+DmC[c].Bal))>>15;
  else
    if(DmC[c].Bal>0)
      DmC[c].LeftVol=(DmC[c].LeftVol*(DM_MAX_VOL-DmC[c].Bal))>>15;
  DmC[c].LeftVol*=2;
  DmC[c].RightVol*=2;
} // end DmUpdateVolume

//******************
// DmPlay(sample)
// starts the sample playing in the background
// returns 0 or channel number
//******************
int DmPlay(DpSample *sample){
  int c;
  
  if(!sample) return 0;  // allows DmPlay(DpLoad...)

  // find an open channel
  for(c=0;c<DmChannels;c++) if(!DmC[c].Active) break;
  if(c==DmChannels) return 0;
  
  // set up the primary channel variables
  DmC[c].Mode=sample->Mode;
  DmC[c].Frequency=sample->Frequency;
  DmC[c].Size=sample->Size;
  DmC[c].Buffer=sample->Buffer;

  // initialize the secondary channel variables
  DmC[c].Vol=DmVol;
  DmC[c].LeftVol=DmLeftVol<<1;
  DmC[c].RightVol=DmRightVol<<1;
  DmC[c].Count=DmC[c].Size;
  DmC[c].Pos=DmC[c].Buffer;
  if(DmC[c].Frequency < DmFrequency){
// step is amount to repeat each time, could be used for volume
// interpolation, however, in the interest of simplicity, ignored for now
//    DmC[c].Fstep=DmFrequency/DmC[c].Frequency;
//    DmC[c].Fadj=DmFrequency%DmC[c].Frequency;
//    DmC[c].Ferr=DmC[c].Frequency;
    DmC[c].Fstep=0;
    DmC[c].Fadj=DmC[c].Frequency;
    DmC[c].Ferr=DmFrequency;
  } else {
    DmC[c].Fstep=DmC[c].Frequency/DmFrequency;  // add to position each time
    DmC[c].Fadj=DmC[c].Frequency%DmFrequency;
    DmC[c].Ferr=DmFrequency;
  } // end if-else
  if(DmC[c].Mode&1) DmC[c].Fstep*=2;  // double step for stereo
  if(DmC[c].Mode&2) DmC[c].Fstep*=2;  // double step for 16 bits

  // and, finally, set active
  DmC[c].Active=1;

  return c+1;
} // end DmPlay

//******************
// DmPlayExt(sample,volume,balance)
// starts the sample playing in the background
// with supplied volume and balance
// returns 0 or channel number
//******************
int DmPlayExt(DpSample *sample,int vol, int bal){
  int c;
  
  if(!sample) return 0;

  // find an open channel
  for(c=0;c<DmChannels;c++) if(!DmC[c].Active) break;
  if(c==DmChannels) return 0;
  
  // set up the primary channel variables
  DmC[c].Mode=sample->Mode;
  DmC[c].Frequency=sample->Frequency;
  DmC[c].Size=sample->Size;
  DmC[c].Buffer=sample->Buffer;

  // initialize the secondary channel variables
  if(vol<0) vol=0;
  if(vol>DM_MAX_VOL*DmChannels) vol=DM_MAX_VOL*DmChannels;
  if(bal<DM_LEFT_BAL) bal=DM_LEFT_BAL;
  if(bal>DM_RIGHT_BAL) bal=DM_RIGHT_BAL;
  DmC[c].Vol=vol;
  DmC[c].Bal=bal;
  DmUpdateVolume(c);  
  DmC[c].Count=DmC[c].Size;
  DmC[c].Pos=DmC[c].Buffer;
  if(DmC[c].Frequency < DmFrequency){
// step is amount to repeat each time, could be used for volume
// interpolation, however, in the interest of simplicity, ignored for now
//    DmC[c].Fstep=DmFrequency/DmC[c].Frequency;
//    DmC[c].Fadj=DmFrequency%DmC[c].Frequency;
//    DmC[c].Ferr=DmC[c].Frequency;
    DmC[c].Fstep=0;
    DmC[c].Fadj=DmC[c].Frequency;
    DmC[c].Ferr=DmFrequency;
  } else {
    DmC[c].Fstep=DmC[c].Frequency/DmFrequency;  // add to position each time
    DmC[c].Fadj=DmC[c].Frequency%DmFrequency;
    DmC[c].Ferr=DmFrequency;
  } // end if-else
  if(DmC[c].Mode&1) DmC[c].Fstep*=2;  // double step for stereo
  if(DmC[c].Mode&2) DmC[c].Fstep*=2;  // double step for 16 bits

  // and, finally, set active
  DmC[c].Active=1;

  return c+1;
} // end DmPlayExt

//******************
// DmStop(channel)
// Stops playback of the specifed channel
// channel 0 stops all channels
//******************
void DmStop(int channel){
  if(channel<1 || channel>DmChannels) return;
  if(!channel)
    for(channel=0;channel<DmChannels;channel++) DmC[channel].Active=0;
   else DmC[channel-1].Active=0;
} // end DmStop

//******************
// DmSetVolume(channel,volume)
// sets volume of channel (DM_REPORT for no change)
// or channel 0=master volume
// returns the current volume
//******************
int DmSetVolume(int channel, int vol){
  if(!channel){
    if(vol>=0 && vol<=DM_MAX_VOL*DmChannels){
      DmVol=vol;
      DmLeftVol=DmVol;
      DmRightVol=DmVol;
      if(DmBal<0) DmRightVol=(DmRightVol*(DM_MAX_VOL+DmBal))>>15;
      else
        if(DmBal>0) DmLeftVol=(DmLeftVol*(DM_MAX_VOL-DmBal))>>15;
      for(channel=0;channel<DmChannels;channel++)
        if(DmC[channel].Active) DmUpdateVolume(channel);
    } // end if
    return DmVol;
  } // end if

  if(!DmC[channel-1].Active) return 0;

  if(vol>=0 && vol<=DM_MAX_VOL*DmChannels){
    DmC[channel-1].Vol=vol;
    DmUpdateVolume(channel-1);
  } // end if
  return DmC[channel-1].Vol;
} // end DmSetVolume

//******************
// DmSetBalance(channel,balance)
// sets balance of channel (DM_REPORT for no change)
// or channel 0=master balance
// returns the current balance
//******************
int DmSetBalance(int channel, int bal){
  if(!channel){
    if(bal>=DM_LEFT_BAL && bal<=DM_RIGHT_BAL){
      DmBal=bal;
      DmLeftVol=DmVol;
      DmRightVol=DmVol;
      if(DmBal<0) DmRightVol=(DmRightVol*(DM_MAX_VOL+DmBal))>>15;
      else
        if(DmBal>0) DmLeftVol=(DmLeftVol*(DM_MAX_VOL-DmBal))>>15;
      for(channel=0;channel<DmChannels;channel++)
        if(DmC[channel].Active) DmUpdateVolume(channel);
    } // end if
    return DmBal;
  } // end if

  if(!DmC[channel-1].Active) return 0;

  if(bal>=DM_LEFT_BAL && bal<=DM_RIGHT_BAL){
    DmC[channel-1].Bal=bal;
    DmUpdateVolume(channel-1);
  } // end if
  return DmC[channel-1].Bal;
} // end DmSetBalance

//******************
// DmSetCounter(channel,count)
// sets counter position for specified channel (DM_REPORT for no change)
// returns the current counter position
// (counts down from sample->Size)
//******************
int DmSetCounter(int channel,int count){
  if(channel<1 || channel>DmChannels || !DmC[channel-1].Active) return 0;
  if(count>=0 && count!=DM_REPORT) DmC[channel-1].Count=count;
  return DmC[channel-1].Count;
} // end DmSetCounter

//******************
// DmPlaying()
// returns 1 if any samples are currently playing
// else 0 
//******************
int DmPlaying(void){
  int c;
  for(c=0;c<DmChannels;c++)
    if(DmC[c].Active) return 1;
  return 0;
} // end DmPlaying


